{-
APCC

Open Source Fixedpoint Model-Checker version 2012b

(C) Copyright Paolo Modesti 2012
(C) Copyright Sebastian Moedersheim 2010,2012
(C) Copyright IBM Corp. 2009

All Rights Reserved.

-}

module Ast where
import Data.List
import Data.Maybe
import Data.Typeable
import Data.Data
import Msg

-- | The type @Protocol@ is the type for the result of the AnB-Parser,
-- i.e. the topmost structure of the Abstract Syntax Tree (AST)
-- | Name, Mappings, Formats and Macros are all optional

-- type Protocol = (Maybe Name, Types, Maybe Mappings, Maybe Formats, Knowledge, Maybe Macros, Actions, Goals)
-- type Protocol = (Name, Types,  Mappings,  Formats, Knowledge,  Macros, Actions, Goals, Abstraction) 
-- Compositional Extension: "Private"  and "Results" sections
type Protocol = (Name, Types,  Mappings,  Formats, Knowledge,  Macros, Actions, Goals, Private, Results, Abstraction) 

type Name=String
-- OLD:type Protocol = (String,Types,Knowledge,Abstraction,Actions,Goals)

----------- Protocol Description --------------

-- Types (see Msg)
-- 
-- | Mappings is a list of tuples each consist of:
-- Mapping: Mapping name, 
-- [Type]: list of parameters types, 
-- and return Type

type Mappings = [(MappingName, [Type], Type)]

type MappingName = Ident
type Formats = [(FormatName, [Type])]
type FormatName = Ident
type Macros = [(MacroName, [Ident], Msg)]
type MacroName = Ident
-- | A knowledge declaration consists of an identifier for a role
-- (constant or variable; should be of type Agent!) and a list of
-- messages that this role initially knows. The variables in that
-- knowledge should always be of type Agent; this is not checked now,
-- however. NEW: Additionally, we have a set of pairs of messages that
-- express inequalities of terms in the instantiation of the
-- knowledge. One can use this for instance to express with A/=B that
-- the roles A and B must be played by different agents.
type Knowledge = ([Aknowledge],[(Msg,Msg)]) --may be to change it to type Knowledge = ([(Ident,[(Msg,Msg)])],[(Msg,Msg)])
type Aknowledge = (Ident,[Msg])

-- | A peer is the endpoint of a channel (i.e. sender or
-- receiver). The identifier is the real name (according to the
-- specification), the bool says whether this agent is
-- pseudonymous. The third is the pseudonym in case the agent is
-- pseudonymous.
-- type Peer = (Ident,Bool,Maybe Msg) -- oldVersion
type Peer = (Ident,Msg, Bool,Maybe Msg) -- I changed this to add channels to ProVerif back-end! The second Message is the name of the peer as a message.

-- | A channel is characterized by a sender and receiver peer, and by
-- a channel type, e.g.  @ ((A,False,Nothing),Secure,(B,True,PKB)) @
type Channel = (Peer,ChannelType,Peer)

-- | An action consists of a channel (sender, channeltype, receiver)
-- and a message being sent. Additionally, there are two optional
-- message terms for modeling zero-knowledge protocols, namely a
-- pattern for the receiver, and a message that the sender must know
data Action = Exchange (Channel,Msg,Maybe Msg,Maybe Msg) 
            | Let (Ident, Msg) 
            | If (Ident, Action, Action)
            | Or (Action, Action)
            | CreateFresh (Channel, (Type, [Ident]))
            deriving (Eq,Show)
--type Let    = (Ident, Msg)
-- | List of actions
type Actions = [Action] --previously!--([Action],[Let]) --

-- | The different supported channel types
data ChannelType = Insecure -- ^ @ -> @ standard channel
                 | Authentic  -- ^ @ *-> @ sender and intended recipient secured
                 | Confidential  -- ^ @ ->* @ recipient secured
                 | Secure   -- ^ @ *->* @ both authentic and confidential
                 | FreshAuthentic  -- ^ @ *->> @ like authentic, but protected against replay
                 | FreshSecure -- ^ @ *->>* @ like secure, but protected against replay
                 | DecFresh -- this is not a channel, but it is for fresh values declaration e.g., A: Nonce X
                 deriving (Eq,Show)

-- | The pre-defined set of goals
data Goal = ChGoal Channel Msg -- ^ Goals that are expressed in channel notation
          | Secret Msg [Peer] -- ^ Standard secrecy goal
          | Authentication Peer Peer Msg -- ^ Standard authentication
                                         -- goal (including replay
                                         -- protection; corresponds to
                                         -- Lowe's injective
                                         -- agreement)
         | WAuthentication Peer Peer Msg -- ^ Weaker form of
                                          -- authentication: no
                                          -- protection against
                                          -- replay. Corresponds to
                                          -- Lowe's non-injective
                                          -- agreement.
      deriving (Eq,Show)

-- | List of goals
type Goals = [Goal]

-- | For the annotation of abstractions: to express that a variable
-- (identifier) should be abstracted into a given term. This
-- represents replacing in the entire description all occurrences of
-- that variable with that term.
type Abstraction = [ (Ident,Msg) ]

--- Compositioanl Extension
type Private = [Msg]
type Results = [(Ident, Ident, Msg)]
--------------------------
instance Show Type where
    show (Agent b1 b2 t) = "Agent " 
    show (PublicKey t) = "PublicKey "
    show (SymmetricKey t) = "SymmetricKey "
    show (PrivateKey t) = "PrivateKey "
    show (Number t) = "Number "
    show (Msg t) = "Msg "
    show (Function t) = "Function "
    show (SeqNumber t) = "SeqNumber "
    show (Untyped t) = "Untyped "
    show (Boolean t) = "Boolean "
