{
{-
APCC - Automated Protocol Composition Checker

Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2014-2015
Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015

Open Source Fixedpoint Model-Checker version 2010a

(C) Copyright IBM Corp. 2009,2010

All Rights Reserved.

-}

module Lexer (Token(..), 
              Ident, 
          AlexPosn(..), alexScanTokens, 
          token_posn) where
import Ast
import Msg


}

%wrapper "posn"

$digit   = 0-9
$alpha   = [a-zA-Z]
$alphaL  = [a-z]
$alphaU  = [A-Z]
$identChar   = [a-zA-Z0-9_]

tokens :-

  $white+            ;
  "#".*                ;
-- "{#" .($white*  .* .$white*)*  "#}"          ;  
  "{#"            {(\ p s -> TBEGINCOMMENT p)}
  "#}"            {(\ p s -> TENDCOMMENT p)}
  "("            { (\ p s -> TOPENP p)  }
  ")"            { (\ p s -> TCLOSEP p) }
  "{|"              { (\ p s -> TOPENSCRYPT p)}
  "|}"              { (\ p s -> TCLOSESCRYPT p)}
  "{"               { (\ p s -> TOPENB p)  }
  "}"            { (\ p s -> TCLOSEB p) }
  ":"            { (\ p s -> TCOLON p)  }
  ";"               { (\ p s -> TSEMICOLON p) }
  "*->*"            { (\ p s -> TSECCH p)  }
  "*->"        { (\ p s -> TAUTHCH p) }
  "->*"             { (\ p s -> TCONFCH p) }
  "->"        { (\ p s -> TINSECCH p)}
  "*->>"            { (\ p s -> TFAUTHCH p) }
  "*->>*"           { (\ p s -> TFSECCH p) }
  "%"            { (\ p s -> TPERCENT p)}
  "!="              { (\ p s -> TUNEQUAL p)}
  "!"               { (\ p s -> TEXCLAM  p)}
  "."            { (\ p s -> TDOT p) }
  ","            { (\ p s -> TCOMMA p) }
  "["            { (\ p s -> TOPENSQB p) } 
  "]"            { (\ p s -> TCLOSESQB p) }
  "="         {(\p s -> TEQUAL p)}            
  "|"         {(\p s -> TVERT p)}

--- Punctuation extension
-- "_" {(\p s -> TUNDERSCORE p)}
  "@" {(\p s -> TAT p)}
--- Sections
  "Protocol"        { (\ p s -> TPROTOCOL p) }
  "Types"        { (\ p s -> TTYPES p)}
    "Mappings"        { (\ p s -> TMAPPINGS p)}
    "Formats"        { (\ p s -> TFORMATS p)}
    "Macros"        { (\ p s -> TMACROS p)}
    "Knowledge"       { (\ p s -> TKNOWLEDGE p)}
  
  "Actions"        { (\ p s -> TACTIONS p)}
    "Main"        { (\ p s -> TMAIN p)}
  "Abstraction" { (\ p s -> TABSTRACTION p)}--%?
  "Goals"        { (\ p s -> TGOALS p)}
------------- Compositional reasoning
  "Results"        { (\ p s -> TRESULTS p)}
  "Private"        { (\ p s -> TPRIVATE p)}
---
  "where"           { (\ p s -> TWHERE p) }
  "authenticates" { (\ p s -> TAUTHENTICATES p)}
  "on"         { (\p s -> TON p)}
  "weakly"         { (\p s -> TWEAKLY p)}
  "secret"      { (\p s -> TSECRET p)}
  "of"     { (\p s -> TOF p)}
  "between"     { (\p s -> TBETWEEN p)}
-- Extension part  :if-then-else-fi,Formats, Mappings....
  "if"         {(\p s -> TIF p)}
  "then"         {(\p s -> TTHEN p)}
  "else"         {(\p s -> TELSE p)}
  "fi"         {(\p s -> TFI p)}
  "or"        {(\p s -> TOR p)}

-- more extension FOR DATABASE SECTION
    "Database"     {(\p s -> TDATABSE p)}      --?
    "?"         {(\p s -> TQUESTION p)}      --TQUERY 
    "+"         {(\p s -> TPLUS p)}         -- TADD
    "-"         {(\p s -> TMINUS  p)}         --OR TDELETE
    "as"         {(\p s -> TAS  p)} 
    "return"     {(\p s -> TRETURN p)}
    "exit"     {(\p s -> TEXIT p)}

--PREDEFINED PRIMITIVES (functions and mappings)
--  "crypt"     {(\p s -> TCRYPT p)}
--  "scrypt"     {(\p s -> TSCRYPT p)}
--  "sign"     {(\p s -> TSIGN p)}
--  "hash"     {(\p s -> THASH p)}
--  "mac"     {(\p s -> TMAC p)}
--  "exp"         {(\p s -> TEXP p)}
--  "mult"         {(\p s -> TMULT p)}
--  "pk"         {(\p s -> TPK p)}
--  "inv"         {(\p s -> TINV p)}
--  "shk"         {(\p s -> TSHK p)}
------- SIGNH
--  "signh"     {(\p s -> TSIGNH p)}
--  "fsignh"     {(\p s -> TSIGNH p)}
-- PREDEFINED PRIMITIVES (datatypes)
 -- "Agent"         {(\p s -> TAGENT p)}
--  "PublicKey"         {(\p s -> TPUBLICKEY p)}
--  "PrivateKey"         {(\p s -> TPRIVATEKEY p)}
--  "Number"         {(\p s -> TNUMBER p)}
--  "SymmetricKey"         {(\p s -> TSYMMETRICKEY p)}
--  "Nonce"         {(\p s -> TNONCE p)}
--  "ImpData"         {(\p s -> TIMPDATA p)}
--  "Msg"         {(\p s -> TMSG p)}
--  "Function"         {(\p s -> TFUNCTION p)}
  
 -- let statement
 "let" {(\p s-> TLET p)}
 --
-- function signature    
    "funsign"  {(\p s -> TFUNSIGN p)}

  "Other"         {(\p s -> TOTHER p)}
  "Agents"         {(\p s -> TAGENTS p)}
  "Nonces"         {(\p s -> TNONCES p)}         --OR TNUMBERS

--Originally there    
  $alpha $identChar* { (\ p s -> TATOM p s) }
  $digit+       { (\ p s -> TATOM p s) }

{

--- type Ident=String

data Token= 
   TATOM AlexPosn Ident    
   | TOPENP AlexPosn
   | TBEGINCOMMENT AlexPosn
   | TENDCOMMENT AlexPosn
   | TCLOSEP AlexPosn
   | TOPENSCRYPT AlexPosn
   | TCLOSESCRYPT AlexPosn
   | TOPENB AlexPosn
   | TCLOSEB AlexPosn
   | TCOLON AlexPosn
   | TSEMICOLON AlexPosn
   | TSECCH AlexPosn
   | TAUTHCH AlexPosn
   | TCONFCH AlexPosn
   | TINSECCH AlexPosn
   | TPERCENT AlexPosn
   | TEXCLAM AlexPosn
   | TDOT AlexPosn
   | TCOMMA AlexPosn
   | TOPENSQB AlexPosn
   | TCLOSESQB AlexPosn
   | TPROTOCOL AlexPosn
   | TKNOWLEDGE AlexPosn
   | TTYPES AlexPosn
   | TACTIONS AlexPosn
   | TABSTRACTION AlexPosn
   | TGOALS AlexPosn
   | TFSECCH AlexPosn
   | TFAUTHCH AlexPosn
   | TAUTHENTICATES AlexPosn
   | TON AlexPosn
   | TWEAKLY AlexPosn
   | TSECRET AlexPosn
   | TBETWEEN AlexPosn
   | TUNEQUAL AlexPosn
   | TWHERE AlexPosn
--Extended part
    | TMAPPINGS AlexPosn
    | TFORMATS AlexPosn
    | TMACROS AlexPosn
-- Compositional Extension
    | TRESULTS AlexPosn
    | TPRIVATE AlexPosn
--    | TCRYPT AlexPosn
--    | TSCRYPT AlexPosn
--    | THASH AlexPosn
--    | TMAC AlexPosn
--    | TEXP AlexPosn
--    | TMULT AlexPosn
--    | TSIGN AlexPosn
--    | TPK AlexPosn
--    | TINV AlexPosn
--    | TSHK AlexPosn
--    | TPUBLICKEY AlexPosn
--    | TPRIVATEKEY AlexPosn
--    | TNUMBER AlexPosn
--    | TSYMMETRICKEY AlexPosn
--    | TNONCE AlexPosn
--    | TIMPDATA AlexPosn
--    | TMSG AlexPosn
    | TLET AlexPosn        
    | TIF AlexPosn
    | TTHEN AlexPosn
    | TELSE AlexPosn
    | TFI AlexPosn
    | TDATABSE AlexPosn
    | TMAIN AlexPosn
    | TQUESTION AlexPosn
    | TPLUS AlexPosn
    | TMINUS  AlexPosn
--    | TDEFINITIONS AlexPosn
    | TAS AlexPosn
--    | TFUNCTION AlexPosn
--    | TASYMMETRIC AlexPosn --TSCRYPT
--    | TSYMMETRIC AlexPosn --TCRYPT
    | TOTHER AlexPosn
    | TAGENTS AlexPosn
    | TNONCES AlexPosn
    | TRETURN AlexPosn
    | TEXIT AlexPosn
    | TEQUAL AlexPosn
    | TVERT AlexPosn
    | TOF AlexPosn
    | TOR AlexPosn
--    | TUNDERSCORE AlexPosn
    | TAT AlexPosn
    | TFUNSIGN AlexPosn
--    | TAGENT AlexPosn
    deriving (Eq,Show)

token_posn (TATOM p _)=p
token_posn (TOPENP p)=p
token_posn (TBEGINCOMMENT p)=p
token_posn (TENDCOMMENT p)=p
token_posn (TCLOSEP p)=p
token_posn (TOPENSCRYPT p)=p
token_posn (TCLOSESCRYPT p)=p
token_posn (TOPENB p)=p
token_posn (TCLOSEB p)=p
token_posn (TCOLON p)=p
token_posn (TSEMICOLON p)=p
token_posn (TSECCH p)=p
token_posn (TAUTHCH p)=p
token_posn (TCONFCH p)=p
token_posn (TINSECCH p)=p
token_posn (TPERCENT p)=p
token_posn (TEXCLAM p)=p
token_posn (TDOT p)=p
token_posn (TCOMMA p)=p
token_posn (TOPENSQB p)=p
token_posn (TCLOSESQB p)=p
token_posn (TPROTOCOL p)=p
token_posn (TKNOWLEDGE p)=p
token_posn (TTYPES p)=p
token_posn (TACTIONS p)=p
token_posn (TABSTRACTION p)=p
token_posn (TGOALS p)=p
token_posn (TFSECCH p)=p
token_posn (TFAUTHCH p)=p
token_posn (TAUTHENTICATES p)=p
token_posn (TWEAKLY p)=p
token_posn (TON p)=p
token_posn (TSECRET p)=p
token_posn (TBETWEEN p)=p
token_posn (TWHERE p)=p
token_posn (TUNEQUAL p)=p

-- Extended part again -- I am boring :(


token_posn (TIF p)=p
token_posn (TTHEN p)=p
token_posn (TELSE p)=p
token_posn (TFI p)=p
token_posn (TDATABSE p)=p
token_posn (TMAIN p)=p
token_posn (TQUESTION p)=p
token_posn (TPLUS p)=p
token_posn (TMINUS  p)=p
--token_posn (TDEFINITIONS p)=p
token_posn (TAS p)=p
--token_posn (TFUNCTION p)=p
--token_posn (THASH p)=p
-- token_posn (TASYMMETRIC p)=p
-- token_posn (TSYMMETRIC p)=p
--token_posn (TMAC p)=p
token_posn (TOTHER p)=p
token_posn (TAGENTS p)=p
token_posn (TNONCES p)=p
token_posn (TRETURN p)=p
token_posn (TEXIT p)=p
token_posn (TEQUAL p)=p
token_posn (TVERT p)=p
-- token_posn (TUNDERSCORE p)=p
token_posn (TAT p)=p
--token_posn (TEXP p)=p
--token_posn (TMULT p)=p
token_posn (TFORMATS p)=p
token_posn (TMAPPINGS p)=p
token_posn (TMACROS p)=p
--- COMPOSITIONAL EXTENSION
token_posn (TRESULTS p)=p
token_posn (TPRIVATE p)=p
--token_posn (TCRYPT p)=p
--token_posn (TSCRYPT p)=p
--token_posn (TMSG p)=p
token_posn (TLET p)=p
--token_posn (TSIGN p)=p
--token_posn (TPK p)=p
--token_posn (TINV p)=p
--token_posn (TSHK p)=p
--token_posn (TPUBLICKEY p)=p
--token_posn (TPRIVATEKEY p)=p
--token_posn (TNUMBER p)=p
--token_posn (TSYMMETRICKEY p)=p
--token_posn (TNONCE p)=p
--token_posn (TIMPDATA p)=p 
token_posn (TOF p)=p
token_posn (TOR p)=p
token_posn (TFUNSIGN p)=p
-- token_posn (TAGENT p)=p
  

}

