{-

  Copyright 2015 Paolo Modesti, School of Computing Science, Newcastle University
  Copyright 2012 Ivana Kojovic, DTU

-}

module MessagePattern (StateType, subTerms, msgPattern,protocolMP, protocolNVST, alphaRenaming,isNonVarSubTerm,zipName) where
import TypeSystem_Defs
import Debug.Trace()
import Msg
import ComposeM (isPrivate,isMap)
import Unification
import Control.Monad.State

-- | Data type for state monad. TermType of the value of the state is the same. Function of the int is to keep counter and do alpha renaming
type StateType = ([(Term,String)],Int)

-- | Function that makes non atomic subterms of single Term and its type, doing alpha-renaming 
subTerms :: (Term,String) -> State StateType StateType
subTerms  ((AMsg (Atom x), protocolName)) = get
-- mappings are considered atomic
subTerms (t@(AMsg (Comp Apply _ xs@(x : _))), protocolName) = if (isMap x) then get else do 
                                                          (state, counter) <- get
                                                          put (state ++ [(alphaRenaming t counter, protocolName)], counter + 1)
                                                          mapM_ (\ x -> subTerms (zipName (AMsg x) protocolName)) xs
                                                          get
                                                            
subTerms (t@(AMsg (Comp op o xs)), protocolName) = if (isPrivate op) then get else do 
                                                          (state, counter) <- get
                                                          put (state ++ [(alphaRenaming t counter, protocolName)], counter + 1)
                                                          mapM (\x-> subTerms (zipName (AMsg x) protocolName)) xs
                                                          get


subTerms (t@(ADes dec m1 m2),protocolName) = do
                                  (state, counter) <- get
                                  put (state ++ [(alphaRenaming t counter, protocolName)], counter + 1)
                                  subTerms (AMsg m1, protocolName) 
                                  subTerms (AMsg m2, protocolName)
                                  get 

subTerms (t@(AVer ver m1 m2),protocolName) = do
                                  (state, counter) <- get
                                  put (state ++ [(alphaRenaming t counter, protocolName)], counter + 1)
                                  subTerms (AMsg m1, protocolName) 
                                  subTerms (AMsg m2, protocolName)
                                  get
                                 
subTerms (t@(TmVar id (VarBind ty)), protocolName) = do
                                  (state, counter) <- get
                                  if isConstant id      -- TO check probably redundant
                                          then put (state ++ [(alphaRenaming t counter, protocolName)], counter + 1)
                                          else put (state,counter)
                                  get 

subTerms  (t@(TmBind id (VarBind ty)), protocolName) = do
                                  (state, counter) <- get
                                  if isConstant id      -- TO check probably redundant
                                          then put (state ++ [(alphaRenaming t counter, protocolName)], counter + 1)
                                          else put (state,counter)
                                  get 

-- | function operates on the single Term to do variable renamings.

msgPattern :: (Term,String) -> State StateType StateType
msgPattern (t,protocolName) = do
                                   (state, counter) <- get
                                   put (state ++ [(alphaRenaming t counter, protocolName)], counter + 1)
                                   get
                                   
-------------------- helpful function

zipName :: Term -> String -> (Term, String)
zipName x name = (x, name)

-- zipName :: Term -> String -> Context -> ((Term,String),TermType)
-- zipName x name ctx = ((x, name), typeof_ts x ctx)

-- zipNames :: [Term] -> String -> Context -> [((Term,String),TermType)]
-- zipNames [] _ _ = []
-- zipNames (x:xs) name ctx = ((x, name), typeof_ts x ctx) : zipNames xs name ctx

-- zipName :: [Term]  String -> [(Term,String)]
-- zipName [] name = []
-- zipName (x:xs) name = (x,name):(zipName xs name)

-- | function is getting Term pattern. It uses msgPattern function

protocolMP :: [(Term,String)] -> ([(Term,String)],Int) -> ([(Term,String)],Int) 
protocolMP [] initState = initState -- fst initstate, then function only gives back one thing
protocolMP (x:xs) initState = let 
                                    newInitState = evalState (msgPattern x) initState -- here msg is msg+protocol name
                              in protocolMP xs newInitState

-- | function compututes non atomic subterms of one protocol
protocolNVST :: [(Term,String)] -> ([(Term,String)],Int) -> ([(Term,String)],Int) 
protocolNVST [] initState = initState 
protocolNVST (x:xs) initState = let 
                                    newInitState = evalState (subTerms x) initState -- msg = msg+protocol name
                                in protocolNVST xs newInitState

-- | alpha renaming function
alphaRenaming :: Term -> Int -> Term
alphaRenaming (AMsg (Atom t)) num | isVariable t = AMsg (Atom (t++"_"++show num))
                                  | otherwise = AMsg (Atom t)
alphaRenaming (AMsg (Comp op o xs)) num = AMsg (Comp op o (map (\ x -> term2msg (alphaRenaming (AMsg x) num)) xs))  
alphaRenaming (ADes des m1 m2) num = ADes des (term2msg (alphaRenaming (AMsg m1) num)) (term2msg(alphaRenaming (AMsg m2) num))   
alphaRenaming (AVer ver m1 m2) num = AVer ver (term2msg (alphaRenaming (AMsg m1) num)) (term2msg(alphaRenaming (AMsg m2) num)) 
alphaRenaming (TmVar id (VarBind t)) num = TmVar (id++"_"++ show num) (VarBind t)
alphaRenaming (TmBind id (VarBind t)) num = TmBind (id++"_"++ show num) (VarBind t)

--
---- | aux function for alphaRenaming
--renameAtomic :: Atomic ->Int -> Atomic -- just because the structure of Mapping
--renameAtomic (Ident x) num = Ident (x++ show num)
--renameAtomic (Mapping id t) num = Mapping id (map (\x -> renameAtomic x num) t )

-- helpful function to see nonatomicsubterms
isNonVarSubTerm:: Term -> Bool -- filter for Map.Map --- additional stuff
isNonVarSubTerm (AMsg (Comp Apply _ ((Atom ('f':'r':'m':rst)):xs))) = False
isNonVarSubTerm (AMsg (Comp Apply _ (x:_))) = not (isMap x)
isNonVarSubTerm (AMsg (Comp op _ _ )) = not (isPrivate op)
isNonVarSubTerm (AMsg (Atom id)) = isConstant id
isNonVarSubTerm _ = True


-- | unification checker between subterms of Term pattern
--checkUnify :: [(Term,String)] -> [(Term,String)] -> Bool  --- main check for theorem
--checkUnify _ [] = True
--checkUnify table (x:arrayMsgtail) = (foldl (\acc a -> if ((unify[(fst x,fst a)] /= Nothing) && (x /= a)) -- TODO type ???? 
--                                        then error ("\n Protocol " ++ show (snd x) ++ " is not format-type safe! \n Unification is possible between terms: "++ show (fst x)++ " and " ++ show (fst a)) 
--                                        else acc) True arrayMsgtail) && checkUnify table arrayMsgtail

--  | first condition of theorem 1 (format type safe)
-- cond1Theorem1 :: [(Term,String)] -> Bool -- MP(P) intersection with variable set is empty?
-- cond1Theorem1 [] = True
-- cond1Theorem1 array = foldl (\acc a -> if (isNonVarSubTerm (fst (fst a))) 
--                                        then acc 
--                                        else do error ("\n Protocol"++show(snd(fst a))++". Element of the Term pattern: "++show(fst(fst a))++" is variable!")) True array 

-- | central function for checking theorem 1: format safe protocol (page 13 in extended version)
-- theoremNum1 :: [(Term,String)] -> Context -> Bool  
-- theoremNum1 array ctx = let nvst = protocolNVST array ctx (Map.empty,0)
--                    in 
--                        if (cond1Theorem1 array) && (checkUnify (fst nvst) (Map.keys (fst nvst)))
--                                then True
--                                else error ("Theorem 1 is violated!") -- never here

-- transformFormat :: ([(Term,Type)],String) -> [((Term,String), Type)]
-- transformFormat ([],_) = []
-- transformFormat (((msg,msgType):array),protocolName) = ((msg, protocolName), msgType) :transformFormat (array, protocolName)

------------------------------------- testing the module