{-

APCC - Automated Protocol Composition Checker

Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2014-2015
Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015

-}

module Preprocessor
where
import Ast
import Msg
import Parser
import Lexer
import GeneratePlainStrands
import System.Environment
import System.IO
import Data.List
import Data.Maybe
-------------------------------------------------------
-- mcrosLetsFrmtsMppngs:
mcrosLetsFrmtsMppngs:: Protocol-> Protocol
mcrosLetsFrmtsMppngs p = prfixFrmatsAndMpping(rplcMcrosAndLets p)
-------------------------------------------------------
--rplcMcrosAndLets: replaces macros and lets then deletes them from the protocol
rplcMcrosAndLets::Protocol -> Protocol
rplcMcrosAndLets p= deleteMcrosAndLets (rplcAllMcros p)
-------------------------------------------------------
-- deleteMcrosAndLets: deletes macros and lets form a protocol : they are not needed any more!
deleteMcrosAndLets p@(nm,tp,mp,fr,kn,mc,ac,gl,pr,rs,ab)= (nm,tp,mp, fr, kn, [], (flterLetsOut ac), gl,pr,rs, ab)
-------------------------------------------------------
-- prfixFrmatsAndMpping: gives formats the prefix:frm and user defined mappings the prefix: mpp
prfixFrmatsAndMpping:: Protocol -> Protocol
prfixFrmatsAndMpping p@(nm,tp,mp,fr,(kn,con),mc,ac,gl,pr,rs,ab)= (nm,tp,mp, fr, kno , mc,act,gl1,pr,rs,ab)
                                                            where kno = ((prfxAllKnow fr mp ud kn),con);
                                                                  act= (prfxActsMsgs fr mp ud ac);
                                                                  ud = userDefFun tp
                                                                  gl1 = prefixFMgoals fr mp ud gl -- paolo added


-- paolo added GOALS MUST BE PREFIXED TOO!
prefixFMgoals fm mp ud (ChGoal ch m1 : goals) = ChGoal ch (prfxOne fm mp ud m1) : prefixFMgoals fm mp ud goals
prefixFMgoals fm mp ud (Secret m1 peers : goals)= Secret (prfxOne fm mp ud m1) peers : prefixFMgoals fm mp ud goals
prefixFMgoals fm mp ud (Authentication p1 p2 m1 : goals) = Authentication p1 p2 (prfxOne fm mp ud m1) : prefixFMgoals fm mp ud goals
prefixFMgoals fm mp ud ((WAuthentication p1 p2 m1):goals) = WAuthentication p1 p2 (prfxOne fm mp ud m1): prefixFMgoals fm mp ud goals
prefixFMgoals fm mp ud []= []

                                                               
--mc, (prfxAct fr mp ac), (prfxAct fr mp gl), ab)
-------------------------------------------------------
-- rplcAllMcros:: Protocol-> Protocol, replaces Macros in a protocol: knowledges, actions and goals
-- Knowledge, Actions , Goals, Abs??
rplcAllMcros::Protocol -> Protocol
rplcAllMcros p@(nm,tp,mp,fr,kn,mc,ac,gl,pr,rs,ab)= (nm,tp,mp,fr, 
                                                (rplcKnosMcros mc kn), 
                                                mc, 
                                                (rplcActsMcros mc ac), 
                                                (rplcGolsMcros (letsToMcros(getLets(rplcActsMcros mc ac))++mc) gl),
                                                pr,rs, 
                                                ab
                                                ) 
-------------------------------------------------------
-- rplcKnosMcros: replaces macros in all knowledges
rplcKnosMcros:: Macros -> Knowledge -> Knowledge
rplcKnosMcros mc (idknows,conds) = ((rplcKnoMcros mc idknows), conds)
-------------------------------------------------------
-- rplcKnoMcros: replaces macros in a knowledges
rplcKnoMcros mc ((id,know):rest) = (id,(rplcMcross mc know)): (rplcKnoMcros mc rest)
rplcKnoMcros mc [] = []
-------------------------------------------------------
-- rplcActsMcros replaces macros in all actions
-- rplcActsMcros:: Macros-> Actions-> Actions -- 
rplcActsMcros mc ((Exchange(ch, m1, m2, m3)):acts)= (Exchange(ch, head (rplcMcross mc [m1]), m2, m3) ):(rplcActsMcros mc acts)
rplcActsMcros mc (l@(Let(id, m1)):acts)= (Let (id, head (rplcMcross mc [m1]))):
                                        (rplcActsMcros (letToMcro (Let (id, head (rplcMcross mc [m1]))):mc) acts)
rplcActsMcros mc ((If(id, act1, act2)):acts)= (If(id, head (rplcActsMcros mc [act1]), head (rplcActsMcros mc [act2]))):
                                                    (rplcActsMcros mc acts)
rplcActsMcros mc ((Or(act1, act2)):acts)= (Or(head (rplcActsMcros mc [act1]), head (rplcActsMcros mc [act2]))):
                                                    (rplcActsMcros mc acts)
-- case Else ( it covers only CreateFresh case!@!@)
rplcActsMcros mc (a:acts) = a:(rplcActsMcros mc acts)
rplcActsMcros mc [] = []
-------------------------------------------------------
-- Alternative Version, the previous one incorporates Lets as macros
-- ->>> rplcActsMcros mc (a:acts)= (rplcActMcros mc a):(rplcActsMcros mc acts)
-- rplcActMcros replaces macros in all actions
-- 1) message exchange : Exchange (Channel,Msg,Maybe Msg,Maybe Msg) (this  version assumes no macros in second and third messages!)
-- However, there is a version that works on all messages (commented next lines)
-- rplcActMcros mc (Exchange(ch, m1, m2, m3)) = Exchange(ch, fromJust (hrplcMcross mc [Just m1]), (hrplcMcross mc [m2]), (hrplcMcross mc [m3]))
-- hrplcMcross mcros [Nothing] = Nothing
-- hrplcMcross mcros [(Just msg)] = Just (head (rplcMcross mcros [msg]))

-- ->>> rplcActMcros mc (Exchange(ch, m1, m2, m3)) = Exchange(ch, head (rplcMcross mc [m1]), m2, m3) 

-- 2) Lets : Let (Ident, Msg) 
-- ->>>  rplcActMcros mc (Let(id, m1)) = Let (id, head (rplcMcross mc [m1]))
--3) If :If (Ident, Action, Action)
-- ->>> rplcActMcros mc (If(id, act1, act2)) = If(id, (rplcActMcros mc act1), (rplcActMcros mc act2))
--4) Or: Or (Action, Action)
-- ->>> rplcActMcros mc (Or(act1, act2)) = Or((rplcActMcros mc act1), (rplcActMcros mc act2))
-------------------------------------------------------
--  rplcGolsMcros replaces macros in all goals
rplcGolsMcros:: Macros-> Goals-> Goals
rplcGolsMcros mc ((ChGoal ch m1):goals)= (ChGoal ch (head(rplcMcross mc [m1]))):(rplcGolsMcros mc goals)
rplcGolsMcros mc ((Secret m1 peers):goals)= (Secret (head(rplcMcross mc [m1])) peers):(rplcGolsMcros mc goals)
rplcGolsMcros mc ((Authentication p1 p2 m1):goals)= (Authentication p1 p2 (head(rplcMcross mc [m1]))):(rplcGolsMcros mc goals)
rplcGolsMcros mc ((WAuthentication p1 p2 m1):goals)= (WAuthentication p1 p2 (head(rplcMcross mc [m1]))):(rplcGolsMcros mc goals)
rplcGolsMcros mc []= []
-------------------------------------------------------
-- get lets from Actions
getLets ((Let(id, m1)):actions)= (Let(id, m1)): getLets actions
getLets (a:actions) =  getLets actions
getLets [] = []
-------------------------------------------------------
-- flterLetsOut: actions without Lets, killing Lets
flterLetsOut ((Let(id, m1)):actions)= flterLetsOut actions
flterLetsOut (a:actions) =  a : (flterLetsOut actions)
flterLetsOut [] = []
-------------------------------------------------------
-- converts a let to a macro
letToMcro (Let(id, m1)) = ("WereLet",[id],m1)
-- converts lets to macros
letsToMcros lets = map letToMcro lets
----------------------------------------------------------
rplcMcross [m] msgs = (rplcMcros m msgs)
rplcMcross [] msgs = msgs
rplcMcross (m:cros) msgs = (rplcMcros m) (rplcMcross cros msgs)
----------------------------------------------------------
rplcMcros mcro [] = []
rplcMcros mcro (x:xs) = (rplcMcro mcro x) : (rplcMcros mcro xs)
----------------------------------------------------------
rplcMcro mcro@(a, [id], msg) (Atom x) = if (x==id && a=="WereLet") then msg else (Atom x)
-- rplcMcro mcro (Atom x) = Atom x
rplcMcro mcro@(a,lst,msg) (Comp Apply opt l@(x:xs)) = if (x == Atom a) then (fndWr a lst  xs msg)    else Comp Apply opt (map (rplcMcro mcro) l)
rplcMcro mcro@(a,lst,msg) (Comp op  opt  l) = Comp op  opt (map (rplcMcro mcro) l)
-- rplcMcro mcro@(a,lst,msg) (Comp Apply (f:b)) = Comp Apply f (map (rplcMcro mcro) b)
----------------------------------------------------------
--fndWr: replace
fndWr nm lst@(x:xs) b msg@(Comp op  opt rst) = Comp op opt  (map (fndWr nm lst b) rst) 
fndWr nm lst@(x:xs) b  msg@(Atom a) =     if ((length lst) /= (length b)) 
                                            then error("Macro "++ nm ++ " has incorrect number of arguments!")
                                        else if (elem a lst) 
                                            then  (b!!(fromJust $ elemIndex a lst))
                                        else msg
                                         -- then fndWr lst (replace msg (elemIndices (Atom x) rst) )  b
----------------------------------------------------------
------------------------
----------------------------------------------------------
------------------------
----------------------------------------------------------
-- prfxAll: to prefix a list of messages (meant for knowledges)
prfxAllKnow fm mp ud [] = []
prfxAllKnow fm mp ud ((id,know):rest) = (id,(prfxAll fm mp ud know)): (prfxAllKnow fm mp ud rest)

-- prfxAll: to prefix a list of messages (meant for a knowledge)
prfxAll fm mp ud []= []
prfxAll fm mp ud (x:xs)= (prfxOne fm mp ud x): (prfxAll fm mp ud xs)

prfxOne fm mp ud (Atom x)= Atom x
prfxOne fm mp ud (Comp Apply  opt ((Atom x):xs))= Comp Apply  opt ((Atom (prfxIt fm mp ud x (length xs)): (map (prfxOne fm mp ud) xs)))
prfxOne fm mp ud (Comp op  opt  xs)= Comp op  opt (map (prfxOne fm mp ud) xs)

----------------------------------------------------------
prfxIt fm mp ud op l = if (elem op (map fst fm)) then 
                        if(l/=ll) then error("The format " ++ op ++ " is given " ++ show(l) ++ " arguments instead of " ++ show(ll)) else "frm"++op
                            else if (elem op (map fst' mp)) then
                                if(l/=lll) then error("The Mapping " ++ op ++ " is given " ++ show(l) ++ " arguments instead of " ++ show(lll)) 
                                    else "mpp"++op
                            -- else if (elem op ud) then "udf"++op            
                        else op
                            where ll= lngthF op fm;
                                  lll= lngthM op mp;
                    -- else if (elem op predefined) then "pdf"++o
                    -- else if (elem op pmappings) then "pdm"++op
                    
                    --error(op++ " is not recognized! is it a function, a mapping, or format?, please specify")
-------------------------------------------------------
lngthF name formatsList= head [(length lst )|(nm,lst)<-formatsList, nm==name]
lngthM name mappingList= head [(length lst )|(nm,lst,ret)<-mappingList, nm==name]
-------------------------------------------------------
-- prfxActsMsgs prefixes formats and mappings in all actions' messages
-- prfxActsMsgs:: -> Actions-> Actions -- 
prfxActsMsgs fr mp ud ((Exchange(ch, m1, m2, m3)):acts)= (Exchange(ch, (prfxOne fr mp ud m1), m2, m3) ):(prfxActsMsgs fr mp ud acts)
prfxActsMsgs fr mp ud (l@(Let(id, m1)):acts)= prfxActsMsgs fr mp ud acts
prfxActsMsgs fr mp ud ((If(id, act1, act2)):acts)= (If(id, head (prfxActsMsgs fr mp ud [act1]), head (prfxActsMsgs fr mp ud [act2]))):
                                                    (prfxActsMsgs fr mp ud acts)
prfxActsMsgs fr mp ud ((Or(act1, act2)):acts)= (Or(head (prfxActsMsgs fr mp ud [act1]), head (prfxActsMsgs fr mp ud [act2]))):
                                                    (prfxActsMsgs fr mp ud acts)
-- this is case else (it only covers the CreateFresh case@!@!) 
prfxActsMsgs fr mp ud (a:acts)= a:(prfxActsMsgs fr mp ud acts)

prfxActsMsgs fr mp ud [] = []
-------------------------------------------------------------------------------------
-- Tuple handlers
-- fst'(a,_,_)=a
-- snd'(_,b,_)=b
-- trd'(_,_,c)=c

--retrieve all protocol identifiers declared in Types
getIdentifiers protocol@(_, t, _, _, _,_, _, _,pr,rs, _) = t
--retrieve all protocol mappings declared in Mappings section
getMappings protocol@(_, _, m, _, _,_, _, _,pr,rs, _) = m
--retrieve all protocol formats  declared in Formats  section
getFormats protocol@(_, _, _, f, _,_, _, _,pr,rs, _) = f
--retrieve all protocol knowledge  declared in Knowledge  section
getKnowledge protocol@(_, _, _, _, k, _,_, _, _,_,_, _) = k
--retrieve all  instantiation constrainds declared in Knowledge  section (wheres)
getWheres protocol@(_, _, _, _,(k,conds), _,_, _, _,_,_) = conds
--retrieve all protocol Macros  declared in Macros  section
getMacros protocol@(_, _, _, _, _,c, _, _,pr,rs, _) = c
--retrieve all protocol Actions  declared in Actions  section
getAllActions protocol@(_, _, _, _, _,_, a, _,pr,rs, _) = a
--retrieve all protocol Goals  declared in Goals  section
getGoals protocol@(_, _, _, _, _,_, _, g,pr,rs, _) = g
--retrieve all protocol Abstractions  declared in Abstractions  section
getAbstractions protocol@(_, _, _, _, _,_, _, _,pr,rs, b) = b
-- getActionsIdentifiers takes an action and returns the message in it! not complete, still IF and OR
-- getActionsIdentifiers ::Action->  Msg
getActionsIdentifiers action@(Exchange (_,m,_,_))= m
getActionsIdentifiers action@(Let(_,n))=n
getActionsIdentifiers action@(If(_,t,_))=getActionsIdentifiers t 
-- getActionsIdentifiers [cf]
getActionsIdentifiers action@(Or(l,_))=getActionsIdentifiers l 
-- : getActionsIdentifiers [l]
getAgents (nm,((Agent _ _ _, ls):xs), mp, fr, kn,mc, ac, gl,pr,rs, ab) = ls
getAgents (nm,(x:xs), mp, fr, kn,mc, ac, gl,pr,rs, ab)  = getAgents (nm,(xs), mp, fr, kn,mc, ac, gl,pr,rs,ab)
getAgents (nm,[], mp, fr, kn,mc, ac, gl,pr,rs, ab)  = []
-----------------------------------------------------------------------

getType (PublicKey anot) = PublicKey --"PublicKey"
getType (PrivateKey anot) = PrivateKey --"PublicKey"
getType (SymmetricKey  anot) = SymmetricKey --"PrivateKey"
getType (Number anot) = Number --"Number"
getType (Msg anot) = Msg --"Msg"
getType (Function  anot) = Function --"Function"
getType (SeqNumber anot) = SeqNumber -- "SeqNumber"
getType (Boolean anot)  = Boolean --"Bool"
getType (Untyped anot)  = Untyped --"Untyped"
getType (Agent b1 b2 anot) = (Agent b1 b2) --"Agent"
-----------------------
-- prnType : print Type
prnType (PublicKey anot) = "PublicKey"
prnType (PrivateKey anot) = "PrivateKey" --"PublicKey"
prnType (SymmetricKey  anot) = "SymmetricKey"
prnType (Number anot) = "Number"
prnType (Msg anot) = "Msg"
prnType (Function  anot) = "Function"
prnType (SeqNumber anot) = "SeqNumber"
prnType (Boolean anot)  = "Bool"
prnType (Untyped anot)  = "Untyped"
prnType (Agent b1 b2 anot) = "Agent"
---------------------------------------------------------------
getAnnot (PublicKey anot) = anot
getAnnot (PrivateKey anot) = anot
getAnnot (SymmetricKey  anot) = anot
getAnnot (Number anot) = anot
getAnnot (Msg anot) = anot
getAnnot (Function anot) = anot
getAnnot (SeqNumber anot) =  anot
getAnnot (Boolean anot)  = anot
getAnnot (Untyped anot)  = anot
getAnnot (Agent b1 b2 anot) = anot

cleanAnnot :: TypeOpts -> [(Ident,Msg)]
cleanAnnot to = [x | y@(Option x) <- to]
                                   
--------------------------------------------
-- getAllActionsIdentifiers takes a list of actions and returns a list of all the messages in them! not complete, since getActionsIdentifiers still missing IF and OR

getAllActionsIdentifiers [] = []
getAllActionsIdentifiers (x:xs) = (getActionsIdentifiers x): (getAllActionsIdentifiers xs)

-- get of msgs into atoms only -}
getAtoms00 [] = []
getAtoms00 (x:xs) = case x of Atom a -> x: (getAtoms00 xs)
                              Comp _ _ rest -> (getAtoms00 (merge rest xs))
                            --  _ -> []


                            

--Merges two lists
merge xs     []     = xs
merge []     ys     = ys
merge (x:xs) (y:ys) = x : y : merge xs ys

{-
getAtoms0 x:xs  
    | x== (Atom a)            = (Atom a): (getAtoms xs)
    | x== (Comp op rest)     = (getAtoms0 rest) : (getAtoms0 xs)
    | otherwise               = []
-}
--getAtoms [] = []
-- getAtoms x@(Atom a)= x 
--getAtoms (Comp op rest)= (getAtoms rest)
--get compMsg@(Comp _ rest):(xs)=    (get rest) : (get xs)

                            
-- getSignatures of msgs into signatures only (functions, mappings, formats, macros....)
-- getSignatures [] = []
-- getSignatures (x:xs) = case x of Atom  a -> (getSignatures xs)
                                -- --Comp Apply (op:rest) -> op: (getSignatures (merge rest xs))
                                 -- Comp op rest -> op: (getSignatures (merge rest xs))
                            -- --  _ -> []


getMacrosAndFormats [] = []
getMacrosAndFormats (x:xs) = case x of Comp Apply  opt (op:rest)-> op: (getMacrosAndFormats (merge rest xs))
                                       Atom _ -> (getMacrosAndFormats xs)
                                       Comp _ _ rest -> (getMacrosAndFormats (merge rest xs))
                                       
                                      
getFormatAndMacros (Atom a)=a
-----------------------------------------------------
userDefFun typs = concat [ b | (a,b)<- typs, a == Function []]
-----------------------------------------------------
iLet a = ("WereLet", [a], (Atom "i"))
-- 
--getSignatures (Atom a)= []
--getSignatures (Comp op rest)=     op: (getSignatures rest)
                            -- if x==(Comp op rest) then op: (getSignatures rest) : (getSignatures xs)
--getSignatures0 ompMsg@(Comp op rest)= op: (getSignatures rest)
-- check if signatures relate to which class () to know if there are anything not related "error"
-- same for atoms, Moreover, relate fresh variables to their owners
-- check for arity of signatures
-- replace lets and macros and produce a new protocol for generating strands
{-

main1 =  do 
    args <- getArgs
    fileName <- return(head args)
    aps <- readFile fileName
    let 
        tokns = alexScanTokens aps
        prot = apspareser tokns
            --acts = getAllActions prot
            --lets = getLets acts
            -- wereLets = letsToMcros lets
            -- newprot= rplcAllMcros prot
        newprot = rplcMcrosAndLets prot
        newprot1 = prfixFrmatsAndMpping newprot
    print newprot1

-}